from typing import Optional, Any
from mypy_extensions import TypedDict

from flask.json import dumps
from werkzeug import Response

RESP_MSGS = {
    401: '未登录',
    403: '未授权',
    404: '不正确的链接地址',
    422: '请求数据校验失败',

    500: '服务器内部错误',
}


class RespType(TypedDict):
    code: str
    message: str
    detail: Any


def gen_response(code: int, message: str = '', detail: Optional[str] = None) \
        -> RespType:
    """Func for generate response body.

    Args:
        code (string, int): Extension to interact with web pages. Default is \
            http response ``status_code`` like 200、404.
        message (string): For popup windows.
        detail (object): For debug, detail server error msg.

    Returns:
        dict: A dict contains all args.

    """
    return {
        'code': str(code),
        'message': message or RESP_MSGS.get(code, '未知错误'),  # type: ignore
        'detail': detail,
    }


class Result(Response):
    """Base json response.
    """
    status = 200  # type: ignore

    def __init__(self, response=None, status=None, headers=None,
                 mimetype='application/json', content_type=None,
                 direct_passthrough=False):
        assert sorted(response.keys()) == ['code', 'detail', 'message'], \
            'Error response, must include keys: code, detail, message'
        super(Result, self).__init__(
            response=dumps(response, indent=0, separators=(',', ':')) + '\n',
            status=status or self.status, headers=headers, mimetype=mimetype,
            content_type=content_type, direct_passthrough=direct_passthrough)


class SuccessResult(Result):
    """Success response. Default status is 200, you can cover it by status arg.
    """
    status = 200

    def __init__(self, message: str = '', code: Optional[int] = None,
                 detail: Any = None, status: Optional[int] = None):
        super(SuccessResult, self).__init__(
            gen_response(code or self.status, message, detail),
            status or self.status)


class FailedResult(Result):
    """Failed response. status always 400.
    """
    status = 400

    def __init__(self, message: str = '', code: Optional[int] = None,
                 detail: Any = None):
        super(FailedResult, self).__init__(
            gen_response(code or self.status, message, detail), self.status)


class UnauthorizedResult(FailedResult):
    status = 401


class ForbiddenResult(FailedResult):
    status = 403


class ValidationErrorResult(FailedResult):
    status = 422


class ServerErrorResult(FailedResult):
    status = 500



from marshmallow import (
    Schema as Schema_, fields, pre_load, post_load, post_dump,
)
from marshmallow_sqlalchemy.schema import ModelSchemaMeta
from flask_marshmallow.sqla import ModelSchema as ModelSchema_
from marshmallow_enum import EnumField


class ORMSchema(ModelSchema_):
    """Base schema for ModelSchema. See `webargs/issues/126
    <https://github.com/sloria/webargs/issues/126>`_.

    Example::

        from hobbit_core.schemas import ORMSchema


        class UserSchema(ORMSchema):

            class Meta:
                model = User
                load_only = ('password')

    ``@use_kwargs(UserSchema())`` use in combination with ``load_only``::

        @bp.route('/users/', methods=['POST'])
        @use_kwargs(UserSchema())
        def create_user(username, password):
            pass
    """

    @post_load()
    def make_instance(self, data):
        return data


class SchemaMixin(object):
    """Add ``id``, ``created_at``, ``updated_at`` fields to schema,
    default ``dump_only=True``.

    Example::

        from marshmallow import Schema

        from hobbit_core.schemas import SchemaMixin

        class UserSchema(Schema, SchemaMixin):
            pass
    """

    id = fields.Int(dump_only=True)
    created_at = fields.DateTime('%Y-%m-%d %H:%M:%S', dump_only=True)
    updated_at = fields.DateTime('%Y-%m-%d %H:%M:%S', dump_only=True)

    class Meta:
        strict = True
        ordered = False
        dateformat = '%Y-%m-%d %H:%M:%S'


class PagedSchema(Schema_):
    """Base schema for list api pagination.

    Example::

        from marshmallow import fields

        from hobbit_core.schemas import PagedSchema

        from . import models
        from .exts import ma


        class UserSchema(ma.ModelSchema):

            class Meta:
                model = models.User


        class PagedUserSchema(PagedSchema):
            items = fields.Nested('UserSchema', many=True)


        paged_user_schemas = PagedUserSchema()
    """

    total = fields.Int()
    page = fields.Int(missing=1, default=1)
    page_size = fields.Int(missing=10, default=10)

    class Meta:
        strict = True


class EnumSetMeta(ModelSchemaMeta):
    """EnumSetMeta is a metaclass that can be used to auto generate load and
    dump func for EnumField.
    """

    @classmethod
    def gen_func(cls, decorator, field_name, enum, verbose=True):

        @decorator
        def wrapper(self, data):
            if data.get(field_name) is None:
                return data

            if decorator is pre_load:
                data[field_name] = enum.load(data[field_name])
            elif decorator is post_dump:
                data[field_name] = enum.dump(data[field_name], verbose)
            else:
                raise Exception(
                    'hobbit_core: decorator `{}` not support'.format(
                        decorator))

            return data
        return wrapper

    def __new__(cls, name, bases, attrs):
        schema = ModelSchemaMeta.__new__(cls, name, tuple(bases), attrs)
        verbose = getattr(schema.Meta, 'verbose', True)

        setattr(schema.Meta, 'dateformat', '%Y-%m-%d %H:%M:%S')

        for field_name, declared in schema._declared_fields.items():
            if not isinstance(declared, EnumField):
                continue

            setattr(schema, 'load_{}'.format(field_name), cls.gen_func(
                pre_load, field_name, declared.enum))
            setattr(schema, 'dump_{}'.format(field_name), cls.gen_func(
                post_dump, field_name, declared.enum, verbose=verbose))

        return schema


class ModelSchema(ORMSchema, SchemaMixin, metaclass=EnumSetMeta):
    """Base ModelSchema for ``class Model(db.SurrogatePK)``.

    * Auto generate load and dump func for EnumField.
    * Auto dump_only for ``id``, ``created_at``, ``updated_at`` fields.
    * Auto set dateformat to ``'%Y-%m-%d %H:%M:%S'``.
    * Auto use verbose for dump EnumField. See ``db.EnumExt``. You can define
      verbose in ``Meta``.

    Example::

        class UserSchema(ModelSchema):
            role = EnumField(RoleEnum)

            class Meta:
                model = User

        data = UserSchema().dump(user).data
        assert data['role'] == {'key': 1, 'label': 'admin', 'value': '管理员'}

    """
    pass



"""
    hobbit_core
    ~~~~~~~~~~~~

    Common utils for flask app.
"""

from flask import Flask
from flask_sqlalchemy import SQLAlchemy


class HobbitManager:
    """Customizable utils management.
    """

    def __init__(self, app=None, db=None, **kwargs):
        """
        app: The Flask application instance.
        """
        self.app = app
        if app is not None:
            self.init_app(app, db, **kwargs)

    def init_app(self, app, db, **kwargs):
        """
        app: The Flask application instance.
        """
        if not isinstance(app, Flask):
            raise TypeError(
                'hobbit_core.HobbitManager.init_app(): '
                'Parameter "app" is an instance of class "{}" '
                'instead of a subclass of class "flask.Flask".'.format(
                    app.__class__.__name__))

        if not isinstance(db, SQLAlchemy):
            raise TypeError('hobbit-core be dependent on SQLAlchemy.')
        self.db = db

        app.config.setdefault('HOBBIT_UPPER_SEQUENCE_NAME', False)

        # Bind hobbit-core to app
        app.hobbit_manager = self



from mypy_extensions import TypedDict
from typing import Union, List

from flask_sqlalchemy import model

from marshmallow import fields
from marshmallow import validate
from webargs.fields import DelimitedList

from .utils import ParamsDict


#: Base params for list view func.
PageParams = ParamsDict(
    page=fields.Int(missing=1, required=False,
                    validate=validate.Range(min=1, max=2**31)),
    page_size=fields.Int(
        missing=10, required=False, validate=validate.Range(min=5, max=100)),
    order_by=DelimitedList(
        fields.String(validate=validate.Regexp(r'^-?[a-zA-Z_]*$')),
        required=False, missing=['-id']),
)
"""Base params for list view func which contains ``page``、``page_size``、\
   ``order_by`` params.

    Example::

        @use_kwargs(PageParams)
        def list_users(page, page_size, order_by):
            pass
"""


class PaginationType(TypedDict):
    items: list
    page: int
    page_size: int
    total: int


def pagination(obj: 'model.DefaultMeta', page: int, page_size: int,
               order_by: Union[str, List[str], None] = 'id', query_exp=None) \
        -> PaginationType:
    """A pagination for sqlalchemy query.

    Args:
        obj (db.Model): Model class like User.
        page (int): Page index.
        page_size (int): Row's count per page.
        order_by (str, list, None): Example: 'id'、['-id', 'column_name'].
        query_exp (flask_sqlalchemy.BaseQuery): Query like \
            ``User.query.filter_by(id=1)``.

    Returns:
        dict: Dict contains ``items``、``page``、``page_size`` and \
            ``total`` fileds.
    """
    if not isinstance(obj, model.DefaultMeta):
        raise Exception('first arg obj must be model.')

    qexp = query_exp or getattr(obj, 'query')

    if order_by:
        if not isinstance(order_by, list):
            order_by = [order_by]

        order_by = [i for i in order_by if i]  # exclude ''

        columns = {i.name for i in obj.__table__.columns}
        diff = {c.lstrip('-') for c in order_by} - columns
        if diff:
            raise Exception(f'columns {diff} not exist in {obj} model')

        order_exp = []
        for column in order_by:
            if column.startswith('-'):
                order_exp.append(getattr(obj, column.lstrip('-')).desc())
            else:
                order_exp.append(getattr(obj, column))
        qexp = qexp.order_by(*order_exp)

    items = qexp.paginate(page, page_size, error_out=False)
    return {
        'items': items.items, 'page': page, 'page_size': page_size,
        'total': items.total,
    }



import os
import random
import string
import pkg_resources

import click

from .handlers import echo
from .handlers.bootstrap import render_project, gen_metadata_by_name, \
    validate_template_path, validate_csv_file, MetaModel, create_models_csv
from . import HobbitCommand, HobbitGroup, CONTEXT_SETTINGS

templates = ['shire', 'rivendell']


@click.group(cls=HobbitGroup, context_settings=CONTEXT_SETTINGS)
def cli():
    pass


def common_options(func):
    func = click.option(
        '-f', '--force', default=False, is_flag=True,
        help='Covered if file exist.')(func)
    func = click.option(
        '-t', '--template', type=click.Choice(templates),
        default='shire', callback=validate_template_path,
        help='Template name.')(func)
    func = click.option(
        '-d', '--dist', type=click.Path(), required=False,
        help='Target path.')(func)
    return func


@cli.command(cls=HobbitCommand)
@click.option('-n', '--name', help='Name of project.', required=True)
@click.option('-p', '--port', help='Port of web server.', required=True,
              type=click.IntRange(1024, 65535))
@common_options
@click.option('--celery/--no-celery', default=False,
              help='Generate celery files or not.')
@click.pass_context
def new(ctx, name, port, dist, template, force, celery):
    """Create a new flask project, render from different template.

    Examples::

        hobbit --echo new -n blog -d /tmp/test -p 1024

    It is recommended to use pipenv to create venv::

        pipenv install -r requirements.txt && pipenv install --dev pytest pytest-cov pytest-env ipython flake8 ipdb
    """  # noqa
    dist = os.getcwd() if dist is None else os.path.abspath(dist)
    ctx.obj['FORCE'] = force
    ctx.obj['CELERY'] = celery
    ctx.obj['JINJIA_CONTEXT'] = {
        'project_name': name,
        'port': port,
        'secret_key': ''.join(random.choice(
            string.ascii_letters) for i in range(38)),
        'version': pkg_resources.get_distribution("hobbit-core").version,
        'celery': celery,
    }

    echo(f'Start init a hobbit project `{name}` to `{dist}`,'
         f' use template {template}')
    render_project(dist, template)
    echo(f'project `{name}` render finished.')


@cli.command(cls=HobbitCommand)
@click.option('-n', '--name', help='Name of feature.', required=True)
@common_options
@click.option('-c', '--csv-path', required=False, type=click.Path(exists=True),
              callback=validate_csv_file,
              help='A csv file that defines some models.')
@click.option('-s', '--tail-slash', default=False, is_flag=True,
              help='Whether to use the trailing slash.')
@click.pass_context
def gen(ctx, name, template, dist, force, csv_path, tail_slash):
    """Generator new feature. Auto gen models/{name}.py, schemas/{name}.py,
    views/{name}.py, services/{name.py}, tests/test_{name}.py etc.
    """
    dist = os.getcwd() if dist is None else os.path.abspath(dist)
    module, _ = gen_metadata_by_name(name)

    if csv_path:
        metadata = {m.name: m for m in MetaModel.csv2model(csv_path)}
    else:
        default = MetaModel.gen_default(name)
        metadata = {default.name: default}

    ctx.obj['FORCE'] = force
    ctx.obj['JINJIA_CONTEXT'] = {
        'name': name,
        'module': module,
        'metadata': metadata,
        'tail_slash': '/' if tail_slash else '',
    }

    render_project(dist, template)


@cli.command(
    cls=HobbitCommand,
    short_help='Create {name}.csv file for gen --csv-path option.',
    help=f"""Create {{name}}.csv file for gen --csv-path option.

    Support type:

        {', '.join(MetaModel.ORM_TYPES + [MetaModel.TYPE_REF])}

    ref type (hobbit_core.db.reference_col) used for ForeignKey.
    """
)
@click.option('-n', '--name', help='Name of csv file.', required=True)
@click.pass_context
def create(ctx, name):
    create_models_csv(name)


cmd_list = [new, gen, create]



import click

from . import HobbitGroup, CONTEXT_SETTINGS
from .handlers.devtools import dev_init


class BootstrapGroup(HobbitGroup):

    @property
    def cmds(self):
        return {func.name: func for func in [init]}


@click.group(cls=BootstrapGroup, context_settings=CONTEXT_SETTINGS)
def dev():
    """Dev tools, more: hobbit dev --help.
    """
    pass


@dev.command()
@click.option('-a', '--all', 'all_', default=False, is_flag=True,
              help='Run all.')
@click.option('--hooks', default=False, is_flag=True, help='Install hooks.')
@click.option('--pipenv', default=False, is_flag=True,
              help='Create virtualenv by pipenv.')
@click.pass_context
def init(ctx, all_, hooks, pipenv):
    """Init dev env: git hooks, pyenv install etc.
    """
    dev_init(all_, hooks, pipenv)



import os
import sys
from subprocess import run, PIPE, STDOUT

import click

from hobbit import ROOT_PATH


def dev_init(all_, hooks, pipenv):
    run('git init', shell=True)

    if all_ or hooks:
        HOOKS_PATH = os.path.join(ROOT_PATH, 'static', 'hooks')
        run(f'cp -r {HOOKS_PATH}/* .git/hooks', shell=True)

    if all_ or pipenv:
        sub = run('which pipenv', shell=True, stdout=PIPE, stderr=STDOUT)
        if sub.returncode != 0:
            click.echo(click.style('cmd pipenv not exist.', fg='red'))
            sys.exit(sub.returncode)
        pipenv_path = sub.stdout.strip().decode('utf8')

        pipenv_cmds = [
            f'{pipenv_path} install --dev pytest pytest-cov pytest-env flake8',
        ]
        if 'requirements.txt' in os.listdir():
            pipenv_cmds.insert(
                0, f'{pipenv_path} install -r requirements.txt --pre')

        cmd = ' && '.join(pipenv_cmds)
        click.echo(click.style(cmd, fg='green'))
        # force pipenv to ignore that environment and create its own instead
        env = os.environ.copy()
        env.update({'PIPENV_IGNORE_VIRTUALENVS': '1'})
        run(cmd, shell=True, env=env)



from collections import Mapping
import inspect
import importlib
import os
import re
import sys
from typing import Any, Dict, List, Optional
from unicodedata import normalize
from distutils.version import LooseVersion

from flask import request
from flask_sqlalchemy import model
import marshmallow
from marshmallow import Schema
from .webargs import use_kwargs as base_use_kwargs, parser


class ParamsDict(dict):
    """Just available update func.

    Example::

        @use_kwargs(PageParams.update({...}))
        def list_users(page, page_size, order_by):
            pass

    """

    def update(self, other=None):
        """Update self by other Mapping and return self.
        """
        ret = ParamsDict(self.copy())
        if other is not None:
            for k, v in other.items() if isinstance(other, Mapping) else other:
                ret[k] = v
        return ret


class dict2object(dict):
    """Dict to fake object that can use getattr.

    Examples::

        In [2]: obj = dict2object({'a': 2, 'c': 3})

        In [3]: obj.a
        Out[3]: 2

        In [4]: obj.c
        Out[4]: 3

    """

    def __getattr__(self, name: str) -> Any:
        if name in self.keys():
            return self[name]
        raise AttributeError('object has no attribute {}'.format(name))

    def __setattr__(self, name: str, value: Any) -> None:
        if not isinstance(name, str):
            raise TypeError('key must be string type.')
        self[name] = value


def secure_filename(filename: str) -> str:
    """Borrowed from werkzeug.utils.secure_filename.

    Pass it a filename and it will return a secure version of it. This
    filename can then safely be stored on a regular file system and passed
    to :func:`os.path.join`.

    On windows systems the function also makes sure that the file is not
    named after one of the special device files.

        >>> secure_filename(u'哈哈.zip')
        '哈哈.zip'
        >>> secure_filename('My cool movie.mov')
        'My_cool_movie.mov'
        >>> secure_filename('../../../etc/passwd')
        'etc_passwd'
        >>> secure_filename(u'i contain cool \xfcml\xe4uts.txt')
        'i_contain_cool_umlauts.txt'

    """

    for sep in os.path.sep, os.path.altsep:
        if sep:
            filename = filename.replace(sep, ' ')

    filename = normalize('NFKD', '_'.join(filename.split()))

    filename_strip_re = re.compile(u'[^A-Za-z0-9\u4e00-\u9fa5_.-]')
    filename = filename_strip_re.sub('', filename).strip('._')

    # on nt a couple of special files are present in each folder.  We
    # have to ensure that the target file is not such a filename.  In
    # this case we prepend an underline
    windows_device_files = (
        'CON', 'AUX', 'COM1', 'COM2', 'COM3', 'COM4', 'LPT1',
        'LPT2', 'LPT3', 'PRN', 'NUL',
    )
    if os.name == 'nt' and filename and \
            filename.split('.')[0].upper() in windows_device_files:
        filename = '_' + filename

    return filename


def _get_init_args(instance, base_class):
    """Get instance's __init__ args and it's value when __call__.
    """
    getargspec = inspect.getfullargspec

    argspec = getargspec(base_class.__init__)
    no_defaults = argspec.args[:-len(argspec.defaults)]
    has_defaults = argspec.args[-len(argspec.defaults):]

    kwargs = {k: getattr(instance, k) for k in no_defaults
              if k != 'self' and hasattr(instance, k)}
    kwargs.update({k: getattr(instance, k) if hasattr(instance, k) else
                   getattr(instance, k, argspec.defaults[i])
                   for i, k in enumerate(has_defaults)})

    assert len(kwargs) == len(argspec.args) - 1, 'exclude `self`'

    return kwargs


def use_kwargs(argmap, schema_kwargs: Optional[Dict] = None, **kwargs: Any):
    """For fix ``Schema(partial=True)`` not work when used with
    ``@webargs.flaskparser.use_kwargs``. More details ``see webargs.core``.

    Args:

        argmap (marshmallow.Schema,dict,callable): Either a
            `marshmallow.Schema`, `dict` of argname ->
            `marshmallow.fields.Field` pairs, or a callable that returns a
            `marshmallow.Schema` instance.
        schema_kwargs (dict): kwargs for argmap.

    Returns:
        dict: A dictionary of parsed arguments.

    """
    schema_kwargs = schema_kwargs or {}

    argmap = parser._get_schema(argmap, request)

    if not (argmap.partial or schema_kwargs.get('partial')):
        return base_use_kwargs(argmap, **kwargs)

    def factory(request):
        argmap_kwargs = _get_init_args(argmap, Schema)
        argmap_kwargs.update(schema_kwargs)

        # force set force_all=False
        only = parser.parse(argmap, request).keys()

        argmap_kwargs.update({
            'partial': False,  # fix missing=None not work
            'only': only or None,
            'context': {"request": request},
        })
        if tuple(LooseVersion(marshmallow.__version__).version)[0] < 3:
            argmap_kwargs['strict'] = True

        return argmap.__class__(**argmap_kwargs)

    return base_use_kwargs(factory, **kwargs)


def import_subs(locals_, modules_only: bool = False) -> List[str]:
    """ Auto import submodules, used in __init__.py.

    Args:

        locals_: `locals()`.
        modules_only: Only collect modules to __all__.

    Examples::

        # app/models/__init__.py
        from hobbit_core.utils import import_subs

        __all__ = import_subs(locals())

    Auto collect Model's subclass, Schema's subclass and instance.
    Others objects must defined in submodule.__all__.
    """
    package = locals_['__package__']
    path = locals_['__path__']
    top_mudule = sys.modules[package]

    all_ = []
    for name in os.listdir(path[0]):
        if not name.endswith(('.py', '.pyc')) or name.startswith('__init__.'):
            continue

        module_name = name.split('.')[0]
        submodule = importlib.import_module(f".{module_name}", package)
        all_.append(module_name)

        if modules_only:
            continue

        if hasattr(submodule, '__all__'):
            for name in getattr(submodule, '__all__'):
                if not isinstance(name, str):
                    raise Exception(f'Invalid object {name} in __all__, '
                                    f'must contain only strings.')
                setattr(top_mudule, name, getattr(submodule, name))
                all_.append(name)
        else:
            for name, obj in submodule.__dict__.items():
                if isinstance(obj, (model.DefaultMeta, Schema)) or \
                        (inspect.isclass(obj) and
                         (issubclass(obj, Schema) or
                          obj.__name__.endswith('Service'))):
                    setattr(top_mudule, name, obj)
                    all_.append(name)
    return all_



from webargs.flaskparser import FlaskParser


class CustomParser(FlaskParser):

    def parse_arg(self, name, field, req, locations=None):
        ret = super().parse_arg(name, field, req, locations=locations)
        if hasattr(ret, 'strip'):
            return ret.strip()
        return ret


parser = CustomParser()
use_args = parser.use_args
use_kwargs = parser.use_kwargs



import os

import click
from click import Command

import inflect

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
ROOT_PATH = os.path.dirname(os.path.abspath(__file__))
inflect_engine = inflect.engine()


class HobbitCommand(Command):

    def format_options(self, ctx, formatter):
        """Writes all the options into the formatter if they exist."""
        # Borrowed from click.MultiCommand
        opts = []
        for param in self.get_params(ctx):
            rv = param.get_help_record(ctx)
            if rv is not None:
                # rewrite for color
                rv = list(rv)
                rv[0] = click.style(rv[0], fg='green')
                opts.append(tuple(rv))

        if opts:
            with formatter.section('Options'):
                formatter.write_dl(opts)


class HobbitGroup(click.Group, HobbitCommand):

    def list_commands(self, ctx):
        return sorted(self.cmds.keys())

    def get_command(self, ctx, cmd_name):
        # Alias startproject to new
        cmd_name = 'new' if cmd_name == 'startproject' else cmd_name
        try:
            return self.cmds[cmd_name]
        except KeyError:
            raise click.UsageError(click.style(
                "cmd not exist: {}\nAvailable ones are: {}".format(
                    cmd_name, ', '.join(self.cmds),
                ), fg='red'))

    @property
    def cmds(self):
        from .bootstrap import cmd_list
        from .devtools import dev
        cmd_list.append(dev)
        return {func.name: func for func in cmd_list}

    def format_options(self, ctx, formatter):
        HobbitCommand.format_options(self, ctx, formatter)
        self.format_commands(ctx, formatter)

    def format_commands(self, ctx, formatter):
        """Extra format methods for multi methods that adds all the commands
        after the options.
        """
        # Borrowed from click.MultiCommand
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            # What is this, the tool lied about a command.  Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue

            commands.append((subcommand, cmd))

        # allow for 3 times the default spacing
        if len(commands):
            limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)

            rows = []
            for subcommand, cmd in commands:
                help = cmd.get_short_help_str(limit)
                rows.append((subcommand, help))

            for i, row in enumerate(rows):  # rewrite for color
                row = list(row)
                row[0] = click.style(row[0], fg='green')
                rows[i] = tuple(row)

            if rows:
                with formatter.section('Commands'):
                    formatter.write_dl(rows)


@click.group(
    cls=HobbitGroup,
    epilog='More details: https://hobbit-core.readthedocs.io/zh/latest/',
    context_settings=CONTEXT_SETTINGS)
@click.version_option()
@click.option('--echo/--no-echo', default=False,
              help='Show the logs of command.')
@click.pass_context
def main(ctx, echo):
    ctx.obj = dict()
    ctx.obj['ECHO'] = echo


if __name__ == '__main__':
    main()



from enum import Enum, EnumMeta
from functools import wraps
import warnings

from mypy_extensions import TypedDict
from typing import Any, Union, List, Dict

from flask import current_app
from sqlalchemy import Integer, Column, ForeignKey, func, DateTime, Sequence
from sqlalchemy.orm.session import Session
from flask_sqlalchemy import DefaultMeta

db = current_app.hobbit_manager.db


class _BaseModel:

    __mapper_args__ = {
        'order_by': 'id',
    }

    def __repr__(self) -> str:
        """You can set label property.

        Returns:
            str: ``<{classname}({pk}:{label!r})>``
        """
        class_name = self.__class__.__name__
        pk = self.id  # type: ignore
        label = getattr(self, "label", "")
        return f'<{class_name}({pk}:{label!r})>'


class SurrogatePK(_BaseModel):
    """A mixin that add ``id``、``created_at`` and ``updated_at`` columns
    to any declarative-mapped class.

    **id**: A surrogate integer 'primary key' column.

    **created_at**: Auto save ``datetime.now()`` when row created.

    **updated_at**: Auto save ``datetime.now()`` when row updated.

    **It is not recommended. See hobbit_core.db.BaseModel.**
    """

    __table_args__ = {'extend_existing': True}  # type: ignore

    id = Column(Integer, primary_key=True)
    created_at = Column(
        DateTime, index=True, nullable=False, server_default=func.now())
    updated_at = Column(
        DateTime, index=True, nullable=False, server_default=func.now(),
        onupdate=func.now())

    def __init_subclass__(cls, **kwargs):
        msg = 'SurrogatePK is Deprecated. See hobbit_core.db.BaseModel.'
        warnings.warn(msg)
        super().__init_subclass__(**kwargs)


class BaseModelMeta(DefaultMeta):

    def __new__(cls, name, bases, attrs):
        if name in ('BaseModelMeta', 'BaseModel'):
            return super().__new__(cls, name, bases, attrs)

        primary_key_name = attrs.get('primary_key_name') or 'id'

        attrs[primary_key_name] = Column(Integer, primary_key=True)
        attrs['created_at'] = Column(
            DateTime, index=True, nullable=False, server_default=func.now())
        attrs['updated_at'] = Column(
            DateTime, index=True, nullable=False, server_default=func.now(),
            onupdate=func.now())

        if db.get_engine(bind=attrs.get('__bind_key__')).name == 'oracle':
            sequence_name = attrs.get('sequence_name') or \
                f'{name}_{primary_key_name}_seq'
            if current_app.config['HOBBIT_UPPER_SEQUENCE_NAME']:
                sequence_name = sequence_name.upper()
            attrs[primary_key_name] = Column(
                Integer,
                Sequence(sequence_name),
                primary_key=True)

        exclude_columns = attrs.get('exclude_columns', [])
        for column in exclude_columns:
            attrs.pop(column)

        return super().__new__(cls, name, bases, attrs)


class BaseModel(_BaseModel, db.Model, metaclass=BaseModelMeta):  # type: ignore # noqa
    """Abstract base model class contains
    ``id``、``created_at`` and ``updated_at`` columns.

    **id**: A surrogate integer 'primary key' column.

    **created_at**: Auto save ``datetime.now()`` when row created.

    **updated_at**: Auto save ``datetime.now()`` when row updated.

    Support **oracle id sequence**, default name is ``{class_name}_id_seq``,
    can changed by ``sequence_name`` and ``HOBBIT_UPPER_SEQUENCE_NAME`` config.
    Default value of app.config['HOBBIT_UPPER_SEQUENCE_NAME'] is False.

    Examples::

        from hobbit_core.db import Column, BaseModel

        class User(BaseModel):
            username = Column(db.String(32), nullable=False, index=True)

        print([i.name for i in User.__table__.columns])
        # ['username', 'id', 'created_at', 'updated_at']

    Can be blocked columns with **exclude_columns**::

        class User(BaseModel):
            exclude_columns = ['created_at', 'updated_at']
            username = Column(db.String(32), nullable=False, index=True)

        print([i.name for i in User.__table__.columns])
        # ['username', 'id']

    Can be changed primary_key's name using **primary_key_name**::

        class User(BaseModel):
            primary_key_name = 'user_id'
            username = Column(db.String(32), nullable=False, index=True)

        print([i.name for i in User.__table__.columns])
        # ['username', 'user_id', 'created_at', 'updated_at']

    Can be changed sequence's name using **sequence_name**
    (worked with oracle)::

        class User(BaseModel):
            sequence_name = 'changed'
            username = Column(db.String(32), nullable=False, index=True)

        # print(User.__table__.columns['id'])
        Column('id', ..., default=Sequence('changed_id_seq'))
    """

    __abstract__ = True


def reference_col(tablename: str, nullable: bool = False, pk_name: str = 'id',
                  onupdate: str = None, ondelete: str = None, **kwargs: Any) \
        -> Column:
    """Column that adds primary key foreign key reference.

    Args:
        tablename (str): Model.__table_name__.
        nullable (bool): Default is False.
        pk_name (str): Primary column's name.
        onupdate (str): If Set, emit ON UPDATE <value> when
          issuing DDL for this constraint. Typical values include CASCADE,
          DELETE and RESTRICT.
        ondelete (str): If set, emit ON DELETE <value> when
          issuing DDL for this constraint. Typical values include CASCADE,
          DELETE and RESTRICT.

    Others:

    See ``sqlalchemy.Column``.

    Examples::

        from sqlalchemy.orm import relationship

        role_id = reference_col('role')
        role = relationship('Role', backref='users', cascade='all, delete')
    """

    return Column(
        ForeignKey('{0}.{1}'.format(tablename, pk_name),
                   onupdate=onupdate, ondelete=ondelete),
        nullable=nullable, **kwargs)


class EnumExtMeta(EnumMeta):

    def __new__(cls, name, bases, attrs):
        obj = super(EnumExtMeta, cls).__new__(cls, name, bases, attrs)

        keys, values = set(), set()
        for _, member in obj.__members__.items():
            member = member.value
            if not isinstance(member, tuple) or len(member) != 2:
                raise TypeError(
                    u'EnumExt member must be tuple type and length equal 2.')
            key, value = member
            if key in keys or value in values:
                raise ValueError(u'duplicate values found: `{}`, please check '
                                 u'key or value.'.format(member))
            keys.add(key)
            values.add(key)

        return obj


class OptType(TypedDict, total=False):
    key: int
    value: str
    label: str


class EnumExt(Enum, metaclass=EnumExtMeta):
    """ Extension for serialize/deserialize sqlalchemy enum field.

    Be sure ``type(key)`` is ``int`` and ``type(value)`` is ``str``
    (``label = (key, value)``).

    Examples::

        class TaskState(EnumExt):
            # label = (key, value)
            CREATED = (0, '新建')
            PENDING = (1, '等待')
            STARTING = (2, '开始')
            RUNNING = (3, '运行中')
            FINISHED = (4, '已完成')
            FAILED = (5, '失败')
    """

    @classmethod
    def strict_dump(cls, label: str, verbose: bool = False) -> Union[int, str]:
        """Get key or value by label.

        Examples::

            TaskState.strict_dump('CREATED')  # 0
            TaskState.strict_dump('CREATED', verbose=True)  # '新建'

        Returns:
            int|str: Key or value, If label not exist, raise ``KeyError``.
        """

        return cls[label].value[1 if verbose else 0]

    @classmethod
    def dump(cls, label: str, verbose: bool = False) -> Dict[str, Any]:
        """Dump one label to option.

        Examples::

            TaskState.dump('CREATED')  # {'key': 0, 'value': '新建'}

        Returns:

            dict: Dict of label's key and value. If label not exist,
            raise ``KeyError``.
        """

        ret = {'key': cls[label].value[0], 'value': cls[label].value[1]}
        if verbose:
            ret.update({'label': label})
        return ret

    @classmethod
    def load(cls, val: Union[int, str]) -> str:  # type: ignore
        """Get label by key or value. Return val when val is label.

        Examples::

            TaskState.load('FINISHED')  # 'FINISHED'
            TaskState.load(4)  # 'FINISHED'
            TaskState.load('新建')  # 'CREATED'

        Returns:
            str|None: Label.
        """

        if val in cls.__members__:
            return val  # type: ignore

        pos = 1 if isinstance(val, str) else 0
        for elem in cls:
            if elem.value[pos] == val:
                return elem.name

    @classmethod
    def to_opts(cls, verbose: bool = False) -> List[Dict[str, Any]]:
        """Enum to options.

        Examples::

            opts = TaskState.to_opts(verbose=True)
            print(opts)

            [{'key': 0, 'label': 'CREATED', 'value': u'新建'}, ...]

        Returns:
            list: List of dict which key is `key`, `value`, label.
        """

        opts = []
        for elem in cls:
            opt = {'key': elem.value[0], 'value': elem.value[1]}
            if verbose:
                opt.update({'label': elem.name})
            opts.append(opt)
        return opts


def transaction(session: Session, nested: bool = False):
    """Auto transaction commit or rollback. This worked with
    ``session.autocommit=False`` (the default behavior of ``flask-sqlalchemy``)
    or ``session.autocommit=True``.
    See more: http://flask-sqlalchemy.pocoo.org/2.3/api/#sessions

    Tips:
        * **Can't** do ``session.commit()`` **in func**, **otherwise raise**
          ``sqlalchemy.exc.ResourceClosedError``: `This transaction is closed`.

        * **Must use the same session in decorator and decorated function**.

        * **We can use nested** if keep top decorated
          by ``@transaction(session, nested=False)`` and all subs decorated
          by ``@transaction(session, nested=True)``.

    Examples::

        from hobbit_core.db import transaction

        from app.exts import db


        @bp.route('/users/', methods=['POST'])
        @transaction(db.session)
        def create(username, password):
            user = User(username=username, password=password)
            db.session.add(user)
            # db.session.commit() error occurred

    We can nested use this decorator. Must set ``nested=True`` otherwise
    raise ``ResourceClosedError`` (session.autocommit=False) or
    raise ``InvalidRequestError`` (session.autocommit=True)::

        @transaction(db.session, nested=True)
        def set_role(user, role):
            user.role = role
            # db.session.commit() error occurred


        @bp.route('/users/', methods=['POST'])
        @transaction(db.session)
        def create(username, password):
            user = User(username=username, password=password)
            db.session.add(user)
            db.session.flush()
            set_role(user, 'admin')
    """

    def wrapper(func):
        @wraps(func)
        def inner(*args, **kwargs):
            if session.autocommit is True and nested is False:
                session.begin()  # start a transaction

            try:
                with session.begin_nested():
                    resp = func(*args, **kwargs)
                if not nested:
                    # commit - begin(), transaction finished
                    session.commit()
            except Exception as e:
                if not nested:
                    session.rollback()
                    session.remove()
                raise e
            return resp
        return inner
    return wrapper



import os
from pathlib import Path, PurePath
from setuptools import setup, find_packages

SUFFIX = '.jinja2'
ROOT_PATH = os.path.split(os.path.abspath(os.path.join(__file__)))[0]
src_path = os.path.join(ROOT_PATH, 'hobbit')


def gen_data(data_root='static'):
    """just for collect static files.
    """
    return [fpath.as_posix() for fpath in Path(
        PurePath(src_path) / data_root).glob(f'**/*{SUFFIX}')]


package_data = gen_data()
# The amount files of `shire[new,gen]` + `rivendell[...]`
assert len(package_data) == 30 + 4 + 31 + 5, \
    'nums of tepl files error, {}'.format(len(package_data))
package_data.append('py.typed')


try:
    import pypandoc
    long_description = pypandoc.convert_file('README.md', 'rst')
except(OSError, ImportError):
    long_description = open('README.md').read()


setup(
    name='hobbit-core',
    version='1.4.4',
    python_requires='>=3.6, <4',
    description='Hobbit - A flask project generator.',
    long_description=long_description,
    author='Legolas Bloom',
    author_email='zhanhsw@gmail.com',
    url='https://github.com/TTWShell/hobbit-core',
    classifiers=[
        'Topic :: Utilities',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'License :: OSI Approved :: MIT License',
    ],
    zip_safe=False,
    packages=find_packages(),
    package_data={'hobbit': package_data},
    install_requires=[],
    extras_require={
        'hobbit_core': [
            'Flask>=1.0.2',
            'flask-marshmallow>=0.9.0',
            'Flask-Migrate>=2.2.1',
            'flask-shell-ipython>=0.3.1',
            'Flask-SQLAlchemy>=2.3.2',
            'marshmallow-enum>=1.4.1',
            'marshmallow-sqlalchemy>=0.14.1',
            'webargs>=5.1.3, <6',
            'mypy-extensions==0.4.1',
            'pyyaml>=4.2b1',
            'marshmallow==v3.0.0rc6',
        ],
        'hobbit': [
            'Click>=6.7',
            'Jinja2>=2.10',
            'inflect>=2.1.0',
        ],
    },
    entry_points={
        'console_scripts': 'hobbit = hobbit:main [hobbit]'
    },
)



from collections import namedtuple
from contextlib import contextmanager
import csv
import os
import re

import click
from jinja2 import Environment, FileSystemLoader, Template

from hobbit import inflect_engine
from . import echo

SUFFIX = '.jinja2'


def regex_replace(s, find, replace):
    """A non-optimal implementation of a regex filter"""
    return re.sub(find, replace, s)


@contextmanager
def chdir(dist):
    cwd = os.getcwd()
    # exist_ok py3 only
    if not os.path.exists(dist):
        echo(f'mkdir\t{dist}')
        os.makedirs(dist)
    os.chdir(dist)
    yield dist
    os.chdir(cwd)


@click.pass_context
def render_project(ctx, dist, tpl_path):
    celery = ctx.obj.get('CELERY')  # gen cmd not have this arg
    context = ctx.obj['JINJIA_CONTEXT']

    jinjia_env = Environment(loader=FileSystemLoader(tpl_path))
    jinjia_env.filters['regex_replace'] = regex_replace

    with chdir(dist):
        for fn in os.listdir(tpl_path):
            origin_path = os.path.join(tpl_path, fn)

            if os.path.isfile(origin_path) and not fn.endswith(SUFFIX):
                continue

            if os.path.isfile(origin_path):
                data = jinjia_env.get_template(fn).render(context)
                fn = Template(fn).render(context)
                render_file(dist, fn[:-len(SUFFIX)], data)
                continue

            if not celery and origin_path.endswith('tasks'):
                continue

            dir_name = Template(fn).render(context)
            render_project(os.path.join(dist, dir_name),
                           os.path.join(tpl_path, fn))


@click.pass_context
def render_file(ctx, dist, fn, data):
    target = os.path.join(dist, fn)
    if os.path.isfile(fn) and not ctx.obj['FORCE']:
        echo(f'exists {target}, ignore ...')
        return

    echo(f'render\t{target} ...')

    with open(fn, 'w') as wf:
        wf.write(data)

    if fn.endswith('.sh'):
        os.chmod(fn, 0o755)


def gen_metadata_by_name(name):
    module = '_'.join(name.split('_')).lower()
    model = ''.join([sub.capitalize() for sub in name.split('_')])

    if module != name or not all(name.split('_')):
        raise click.UsageError(click.style(
            f'Name `{name}` should be lowercase, with words separated by '
            f'underscores as necessary to improve readability.', fg='red'))

    return module, model


class MetaModel:
    Column = namedtuple('Column', [
        'field', 'type', 'type_arg', 'is_unique', 'is_index', 'is_null',
        'doc', 'test'])

    """ Gen types
    from sqlalchemy import types
    from sqlalchemy.sql.visitors import VisitableType

    [
        k for k, v in types.__dict__.items()
        if isinstance(v, VisitableType) and k != k.upper()
        and not k.startswith('_')]
    """
    ORM_TYPES = [
        'BigInteger',
        'Binary',
        'Boolean',
        'Date',
        'DateTime',
        'Enum',
        'Float',
        'Integer',
        'Interval',
        'LargeBinary',
        'MatchType',
        'NullType',
        'Numeric',
        'PickleType',
        'SmallInteger',
        'String',
        'Text',
        'Time',
        'Unicode',
        'UnicodeText',

        'JSON',
    ]
    ORM_TYPE_MAPS = {t.lower(): t for t in ORM_TYPES}
    TYPE_REF = 'ref'

    @classmethod
    def gen_model(cls, name=''):
        module, model = gen_metadata_by_name(name)

        class Model:
            name = model
            singular = module
            plural = inflect_engine.plural(module)
            columns = []
            refs = []

        return Model

    @classmethod
    def cleaning_test_value(cls, type_, value):
        if type_ in ['BigInteger', 'Float', 'Integer', 'SmallInteger']:
            return value
        if type_ == 'Boolean':
            return value.capitalize()
        return f"'{value}'"

    @classmethod
    def row_processing(cls, row):
        column = cls.Column(*row)

        gen_metadata_by_name(column.field)  # validate the field name

        type_ = column.type.lower()
        assert type_ in cls.ORM_TYPE_MAPS or type_ == cls.TYPE_REF, \
            f'column type err: cannot be `{column.type}`'

        type_ = cls.ORM_TYPE_MAPS.get(type_) or cls.TYPE_REF
        is_unique = True if column.is_unique else False
        is_index = True if column.is_index else False
        is_null = True if column.is_null else False
        test = cls.cleaning_test_value(type_, column.test)

        column = column._replace(
            type=type_, is_null=is_null,
            is_unique=is_unique, is_index=is_index, test=test)

        return column

    @classmethod
    def csv2model(cls, csv_path):
        with open(csv_path) as csvfile:
            spamreader = csv.reader(csvfile)

            Model = None
            for row in spamreader:
                if spamreader.line_num == 1:
                    continue
                if len(row) == 1 or row[0] == ''.join(row):
                    if Model is not None:
                        yield Model
                    Model = cls.gen_model(row[0])
                elif len(row) == 8:
                    row = cls.row_processing(row)
                    if row.type != cls.TYPE_REF:
                        Model.columns.append(row)
                    else:
                        Model.refs.append(row)
                else:
                    raise click.UsageError(
                        click.style(f'csv file err: `{row}`.', fg='red'))
            yield Model

    @classmethod
    def gen_default(cls, name):
        model = cls.gen_model(name)
        model.columns.append(cls.row_processing([
            'username', 'String', 20, '', 'index', '', '用户名', 'test']))
        return model


def validate_template_path(ctx, param, value):
    from hobbit import ROOT_PATH
    dir = 'feature' if ctx.command.name == 'gen' else 'bootstrap'
    tpl_path = os.path.join(ROOT_PATH, 'static', dir, value)

    if not os.path.exists(tpl_path):
        raise click.UsageError(
            click.style(f'Tpl `{value}` not exists.', fg='red'))

    return tpl_path


def validate_csv_file(ctx, param, value):
    if value is None:
        return value

    with open(value) as csvfile:
        spamreader = csv.reader(csvfile)
        for row in spamreader:
            if len(row) == 1 or row[0] == ''.join(row) or len(row) == 8:
                continue
            else:
                raise click.UsageError(
                    click.style(f'csv file err: `{row}`.', fg='red'))
    return value


def create_models_csv(name):
    with open(name, 'w') as csvfile:
        fieldnames = MetaModel.Column._fields
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        writer.writeheader()
        writer.writerow({'field': 'user'})
        writer.writerow({
            'field': 'username', 'type': 'string', 'type_arg': 16,
            'is_unique': True, 'doc': '用户名', 'test': 'admin',
        })
        writer.writerow({
            'field': 'age', 'type': 'integer', 'is_null': True,
            'doc': '年龄', 'test': 35,
        })
        writer.writerow({
            'field': 'is_admin', 'type': 'boolean',
            'doc': '是否管理员', 'test': True,
        })
        writer.writerow({
            'field': 'role_id', 'type': 'ref', 'type_arg': 'roles',
            'doc': '角色id', 'is_null': True,
        })
        writer.writerow({'field': 'role'})
        writer.writerow({
            'field': 'name', 'type': 'string', 'type_arg': 16,
            'is_unique': True, 'doc': '角色名', 'test': 'admin',
        })
